// Copyright (C) 2009 - Digiteo - Michael Baudin
// Copyright (C) INRIA - Farid BELAHCENE
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt

//
// number_factor --
//   Returns the prime factors of n, as a row matrix
//   If n=0, 1, 2 or 3, then returns n and status = %t.
//   * In case of success, all factors are returned in the result row matrix,
//     and success = %t.
//   * In case of failure, success = %f and the factor which were found
//     are returned in result. 
//     In this case, result = n if no factor has been found.
// Arguments :
//  n : a real scalar (positive)
//  result : a row vector of primes numbers
//  result : a row matrix containing the divisors
//  status : %t if the factorization is a sucecss; 
//           %f if the factorization is a failure
// Note
//   With Pollard's rho method, failure is obtained either when the 
//   number is prime or composite.
//
function [ result , status ] = number_factor ( n , method )
  if (~isdef('method','local')) then
    method = "fasttrialdivision";
  end
  if ( type(n)<>1 ) then
    error(msprintf(gettext("%s: Wrong type for input argument #%d: A real expected.\n"),"number_factor" , 2 ));
  end
  if ( n<0 ) then
    errmsg = sprintf ( gettext ( "%s: Number %d is not positive" ) , "number_factor" , n )
    error(errmsg)
  end
  if ( n<0 | n-floor(n)<>0 )
        error(gettext ( "%s: Input argument %s must be a positive integer") , "number_factor" , string(n) )
  end
  if ( or(n==[0 1 2 3]) ) then 
    result = n
    status = %t
    return
  end
  imax = number_cget ( "-imax" )
  select method
  case "trialdivision"
    [ result , status ] = factor_trialdivision ( n , imax )
  case "fasttrialdivision"
    [ result , status ] = factor_fasttd ( n )
  case "memorylesstd"
    [ result , status ] = factor_memorylesstd ( n , imax )
  case "fermat"
    [ result , status ] = factor_fermat ( n , imax )
  case "pollard-brent"
    [ result , status ] = factor_pollard ( n , "brent" , imax )
  case "pollard-floyd"
    [ result , status ] = factor_pollard ( n , "floyd" , imax )
  else
    errmsg = sprintf ( gettext ( "%s: Unknown method %s" ) , "number_factor" , method )
    error(errmsg)
  end
endfunction
// 
// factor_fasttd --
//   Uses a vectorized trial division algorithm.
//   Uses an array with size sqrt(n).
// Author : F.Belahcene
//
function [ result , status ] = factor_fasttd ( x )
  s = sqrt ( n )
  number_log ( sprintf ( "Computing primes less than %s\n" , string ( s ) ) )
  xprimefact = number_primes(s)
  result = []
  status = %f
  for i = 1 : imax
    number_log ( sprintf ( "Loop #%d/%d, x=%s\n" , i , imax , string ( x ) ) )
    if ( x==1 ) then
      number_log ( sprintf ( "Found x=1, done.\n" ) )
      status = %t
      break
    end
    xdiv = x ./ xprimefact
    xprimefact_ind = find(xdiv-floor(xdiv)==0)
    if ( xprimefact_ind <> [] )
      xprimefact = xprimefact ( xprimefact_ind )
      number_log ( sprintf ( "Found factors %s \n" , strcat ( string ( xprimefact ) , " " ) ) )
      result = [xprimefact result]
      x = x/prod(xprimefact)
    else
      number_log ( sprintf ( "Found no divisor, done.\n" ) )
      result=[x result]
      status = %t
      break
    end
  end
  if ( ~status ) then
    number_log ( sprintf ( "Failed to factor completely, x = %s.\n" , string ( x ) ) )
    result ( 1 , $ + 1 ) = x
  end
  result = gsort(result,'c','i')
endfunction

//
// factorization_trialdivision --
//   Returns the list of prime factors of n
//   Uses the Trial Division method, which is known as the 
//   simplest, and the least efficient method.
// Arguments:
//   n : the positive integer number to factor
//   imax : the maximum number of iterations in the search for a divisor
// Notes:
//   This is Algorithm A, "Factoring by division", section 4.5.4 in Knuth's TAO
//
function [ result , status ] = factor_trialdivision ( n , imax )
    result = []
    // Get all primes up to sqrt(n)
    s = ceil ( sqrt ( n ) )
    number_log ( sprintf ( "Computing primes less than %s\n" , string ( s ) ) )
    prms = number_primes ( s )
    if ( prms( $ ) < s ) then
      // Find a prime greater or equal to s
      p = prms( $ )
      while ( %t )
        p = p + 2
        isprime = number_isprime(p)
        if ( isprime ) then
          prms ( $ + 1 ) = p
          break
        end
      end
    end
    // Divide by prime until the remainder is not zero
    u = n
    // ip is the index of the current prime
    ip = 1
    p = prms ( ip )
    nbp = size ( prms , "*" )
    status = %f
    for i = 1 : imax
      number_log ( sprintf ( "Loop #%d/%d, u=%s, trying p = %s\n" , i , imax , string ( u ) , string ( p ) ) )
      if ( u == 1 ) then
        number_log ( sprintf ( "Found u=1, done.\n" ) )
        status = %t
        break
      end
      r = modulo ( u , p )
      if ( r==0 ) then
        number_log ( sprintf ( "Found divisor p=%s.\n" , string ( p ) ) )
        result ( 1 , $ + 1 ) = p
        u = u / p
      else
        // p does not divide u
        if ( ip < nbp ) then
          // There are primes left to test
          ip = ip + 1
          p = prms ( ip )
          number_log ( sprintf ( "Trying new divisor #%d, p=%s.\n" , ip , string ( p ) ) )
        else
          // There are no primes left to test : the last number n is a prime
          number_log ( sprintf ( "No prime left to test, done.\n" ) )
          result ( 1 , $ + 1 ) = u
          status = %t
          break
        end
      end
    end
    if ( u == 1 ) then
      number_log ( sprintf ( "Found u=1, done.\n" ) )
      status = %t
      return
    end
    if ( ~status ) then
      number_log ( sprintf ( "Failed to factor completely, u = %s.\n" , string ( u ) ) )
      result ( 1 , $ + 1 ) = u
    end
endfunction
//
// factor_fermat --
//   Returns the list of prime factors of n
//   Uses the Fermat's method.
//   The first number in the list is allways 1.
// Arguments:
//   n : the positive integer number to factor
//   imax : the maximum number of iterations in the search for a divisor
//
function [ result , status ] = factor_fermat ( n , imax )
  // Factors by two until n is odd
  [ twofactors , status ] = factor_bytwo ( n , imax )
  if ( ~status ) then
    result = []
    return
  end
  result = twofactors ( 1 : $ - 1 )
  u = twofactors ( $ )
  if ( u==1 ) then
    return
  end
  // Compute factors for odd number u
  [ otherfactors , status ] = factor_fermatdo ( u , imax )
  result = [ result otherfactors ]
  result = gsort ( result , "g" , "i" )
endfunction
//
// factor_fermatdo --
//   Performs the algorithm recursively on the 
//   given odd number.
//   Returns only the factors which are different from 1.
//   This is Algorithm C, "Factoring by addition and substraction", 
//   section 4.5.4 in Knuth's TAO
// Arguments:
//   n : the positive, odd, integer number to factor 
//   imax : the maximum number of iterations in the search for a divisor
// Note :
// The algorithm is very slow to factor 33554431.
// factor_fermatdo ( 33554431 )
//
function [ result , status ] = factor_fermatdo ( n , imax )
    [ p1 , p2 , status ] = factor_fermatcore ( n , imax )
    if ( ~status ) then
      result = n
      return
    end
    isprime = ( ( p1==1 ) & ( p2==n ) )
    // Factor p1 and p2 recursively, if required
    result = []
    if ( isprime ) then
      // If n is prime, there is no need for recursive calls
      // since n = 1 x n (where the 1 factor is discarded)
      result ( 1 , $ + 1 ) = p2
    else
      // Factorize p1
      [ otherfactors1 , status ] = factor_fermatdo ( p1 , imax )
      if ( ~status ) then
        result = [p1 p2]
        return
      end
      // Be p1 prime or composite, add its factors.
      result = [ result otherfactors1 ]
      // Factorize p2
      [ otherfactors2 , status ] = factor_fermatdo ( p2 , imax )
      if ( ~status ) then
        result = [p1 p2]
        return
      end
      // Be p2 prime or composite, add its factors.
      result = [ result otherfactors2 ]
    end
    status = %t
endfunction
//
// factor_fermatcore --
//   Given an odd integer, returns two 
//   factors p1 and p2 so that n = p1 x p2.
//   The factors may be non prime.
//   This is Algorithm C, "Factoring by addition and substraction", 
//   section 4.5.4 in Knuth's TAO
//   Returns p1 = 0 and p2 = 0 if the procedure fails 
//   before imax iterations.
// Arguments:
//   n : the positive, odd, integer number to factor 
//   imax : the maximum number of iterations in the search for a divisor
// Note :
// * The algorithm is very slow to factor 33554431.
//   factor_fermatdo ( 33554431 )
// * factor_fermatcore is faster than factor_fermatcore2
//
function [ p1 , p2 , status ] = factor_fermatcore ( n , imax )
    a = ceil(sqrt(n))
    status = %f
    for i = 1 : imax
      number_log ( sprintf ( "Loop #%d/%d, a = %d.\n" , i , imax , a ) )
      b2 = a * a - n
      // Computes if b is a square
      b = ceil(sqrt(b2))
      //mprintf ( "a=%d, b=%d, n=%d, a^2 - b^2=%d, |n-(a^2-b^2)|=%d\n", a, b, n, a^2 - b^2 , abs(n-a^2+b^2))
      if ( b^2 == b2 ) then
        // b2 is a square so that b2 = b^2 and 
        // n = a^2 - b^2
        number_log ( sprintf ( "Found b=%d, break.\n" , b ) )
        status = %t
        break
      end
      a = a + 1
    end
  // n = (a - b) * (a + b)
  if ( status ) then
    p1 = a - b
    p2 = a + b
    number_log ( sprintf ( "Found divisors p1=%d, p2=%d, done.\n" , p1 , p2 ) )
  else
    number_log ( sprintf ( "Found no divisor for n=%d.\n" , n ) )
    p1 = 0
    p2 = 0
  end
endfunction
//
// factor_fermatcore2 --
//   Given an odd integer, returns two 
//   factors p1 and p2 so that n = p1 x p2.
//   The factors may be non prime.
//   This is Algorithm C, "Factoring by addition and substraction", 
//   section 4.5.4 in Knuth's TAO
//   This is an implementation as close as possible to Knuth's
//   algorithm
//   Returns p1 = 0 and p2 = 0 if the procedure fails 
//   before imax iterations.
// Arguments:
//   n : the positive, odd, integer number to factor 
//   imax : the maximum number of iterations in the search for a divisor
// Note :
// The algorithm is very slow to factor 33554431.
// factor_fermatdo ( 33554431 )
//
function [ p1 , p2 , status ] = factor_fermatcore2 ( n , imax )
    step = 1
    status = %f
    for i = 1 : imax
      select step
      case 1
        // Initialize
        xp = 2 * ceil(sqrt(n)) + 1
        yp = 1
        r = (ceil(sqrt(n)))^2 - n
        step = 2
      case 2
        // Test r
        if ( r <= 0 ) then
          step = 4
        else
          step = 3
        end
      case 3
        // Step y
        r = r - yp
        yp = yp + 2
        step = 2
      case 4
        // Done ?
        if ( r == 0 ) then
          status = %t
          break
        else
          step = 5
        end
      case 5
        // Step x
        r = r + xp
        xp = xp + 2
        step = 3
      end
    end
  if ( status ) then
    p1 = ( xp - yp ) / 2
    p2 = ( xp + yp - 2 ) / 2
  else
    p1 = 0
    p2 = 0
  end
endfunction
//
// floydprng --
//   The pseudo-random number generator suggested by Pollard.
// Arguments:
//   x : the previous element in the sequence
//   n : the modulo of the prng
//
function y = floydprng ( x , n )
    xm1 = x * x - 1
    y = modulo ( xm1 , n )
endfunction
//
// brentprng --
//   The pseudo-random number generator suggested by Brent.
// Arguments:
//   x : the previous element in the sequence
//   n : the modulo of the prng
//
function y = brentprng ( x , n )
    xm1 = x * x + 3
    y = modulo ( xm1 , n )
endfunction
//
// floyd --
//   Performs Floyd's algorithm with Pollard's condition.
//   Returns one prime factor of n. 
//   If no prime factor is found, returns 0 and status = %f.
//   Caution ! This algorithm may fail !
// Arguments:
//   seed : the seed for the pseudo random number generator
//   n : the positive integer number to factor
//   imax : the maximum number of iterations in the search for a divisor
//   rng : the random number generator, "brent", "floyd"
// Note
//   This is Algorithm B, "Monte Carlo Factorization", 
//   section 4.5.4 in Knuth's TAO
//
function [ result , status ] = floyd ( n , seed , imax , rng )
    x = seed
    y = seed
    j = 0
    g = n
    status = %f
    select rng 
    case "brent"
    for i = 1 : imax
      number_log ( sprintf ( "Loop #%d/%d, x=%d, y=%d.\n" , i , imax , x , y ) )
      j = j + 1
      x = brentprng ( x , n )
      y = brentprng ( y , n )
      y = brentprng ( y , n )
      s = abs ( x - y )
      g = number_gcd ( s , n )
      if ( g > 1 ) then
        break
      elseif ( g == n ) then
        break
      end
    end
    case "floyd"
    for i = 1 : imax
      number_log ( sprintf ( "Loop #%d/%d, x=%d, y=%d.\n" , i , imax , x , y ) )
      j = j + 1
      x = floydprng ( x , n )
      y = floydprng ( y , n )
      y = floydprng ( y , n )
      s = abs ( x - y )
      g = number_gcd ( s , n )
      if ( g > 1 ) then
        break
      elseif ( g == n ) then
        break
      end
    end
    else
      error ( msprintf ( gettext ( "%s: Unknown random number generator %s" ) , "floyd" , rng ) )
    end
    if ( g == n ) then
      number_log ( sprintf ( "Found no factor, done.\n" ) )
      result = 0
      status = %f
    else
      if ( ( g == 1 ) & ( n<>1 ) ) then        
        number_log ( sprintf ( "Found no factor, done.\n" ) )
        result = 0
        status = %f
      else
        number_log ( sprintf ( "Found factor g = %d, done.\n" , g ) )
        result = g
        status = %t
      end
    end
endfunction
//
// floydloop --
//   Performs a loop over Floyd's algorithm with several seeds.
//   Returns the first factor which has been found, or 0 if no 
//   factor has been found yet.
// Arguments:
//   n : the positive integer number to factor
//   seedlist : the list of seed to use (default = [ 0 1 2 ]
//   imax : the maximum number of iterations in the search for a 
//     divisor (default 100000)
//   rng : the random number generator, "brent", "floyd"
//
function [ d , status ] = floydloop ( n , seedlist , imax  , rng )
    status = %f
    for seed = seedlist
    number_log ( sprintf ( "Trying seed=%d\n" , seed ) )
      [ d , status ]= floyd ( n , seed , imax , rng )
      if ( d <> 0 ) then
        status = %t
        break
      end
    end
endfunction
//
// factor_pollard --
//   Returns the list of prime factors of n
//   Uses the Pollard method.
//   Returns 0 if the method fails to compute the factors.
// Arguments:
//   n : the positive integer number to factor
//   rng : the random number generator, "brent", "floyd"
//   imax : maximum number of loops in Floyd's loop.
//
function [ result , status ] = factor_pollard ( n , rng , imax )
    result = []
    // Factors by two until n is odd
    [ twofactors , status ] = factor_bytwo ( n , imax )
    if ( ~status ) then
      result = []
      return
    end
    twolist = twofactors ( 1 : $-1 )
    result  = [ result twolist ]
    u = twofactors ( $ )
    // Compute factors for odd number u
    if ( u <> 1 ) then
      [ otherfactors , status ] = factor_pollarddo ( u , rng , imax )
      result = [ result otherfactors ]
      result = gsort ( result , "g" , "i" )
    end
endfunction
//
// factor_pollarddo --
//   Returns the list of prime factors of the odd number n
//   Uses the Pollard method.
//   Returns status = %f if the method fails to compute the factors.
// Arguments:
//   n : the positive integer number to factor
//   rng : the random number generator, "brent", "floyd"
//   imax : maximum number of loops in Floyd's loop.
//
function [ result , status ] = factor_pollarddo ( n , rng , imax )
    number_log ( sprintf ( "factor_pollarddo: n=%d\n" , n ) )
    if ( u == 1 ) then
      result = 1
      status = %t
      return
    end
    result = []
    seedlist = [ 0 1 2 ];
    [ d , status ] = floydloop ( n , seedlist , imax , rng )
    if ( ~status ) then
      // The Pollard method has failed ; two cases are possible :
      // * the number n is prime
      // * the number n is composite, but Pollard's algorithm cannot
      // factor it.
      // In both case, we add n to the list and exit without
      // recursive calls.
      number_log ( sprintf ( "Found no divisor, done.\n" ) )
      result ( 1 , $ + 1 ) = n
      return
    end
    number_log ( sprintf ( "Found divisor: %d\n" , d ) )
    // The pollard method has found a factor.
    // But that factor d may be composite itself : factor d
    [ otherfactors , status ] = factor_pollarddo ( d , rng , imax )
    // We do not care about the status here: be it failing or sucessfull,
    // take the factors "as is".
    result = [ result otherfactors ]
    u = n / d
    if ( u <> 1 ) then
      // Factorize u
      [ otherfactors , status ] = factor_pollarddo ( u , rng , imax )
      // Be u prime or composite, add its factors.
      result = [ result otherfactors ]
    end
endfunction
//
// factor_bytwo --
//   Divide the given n by 2 until the number becomes odd.
//   Returns the list of 2 which were found plus the odd
//   number at the end of the list.
// Arguments:
//   n : the positive integer number to factor
//
function [ result , status ] = factor_bytwo ( n , imax )
    result = []
    // Divide by 2 until the remainder is not zero
    u = n
    status = %f
    for i = 1 : imax
      number_log ( sprintf ( "Loop #%d/%d, u=%d\n" , i , imax , u ) )
      if ( modulo ( u , 2 ) == 0 ) then
        number_log ( sprintf ( "Found divisor %d\n" , 2 ) )
        result ( 1 , $ + 1 ) = 2
        u  = u / 2
      else 
        // Number is odd
        number_log ( sprintf ( "No divisor left, done.\n" ) )
        status = %t
        break
      end
    end
    result ( 1 , $ + 1 ) = u
endfunction
//
// factorization_memorylesstd --
//   Uses the Trial Division method, without any list of primes.
// Arguments:
//   n : the positive integer number to factor
//   imax : the maximum number of iterations in the search for a divisor
// Note:
//   This is Algorithm A, "Factoring by division", section 4.5.4 in Knuth's TAO
// Divide by prime until the remainder is not zero
// * add2 is a boolean which allows to compute the next prime
//   The following primes are tested in order:
//   2 3 5 7 11 13 17 ...
//   After the prime p=5 has been used, we add 2, then 4, then 2.
// * passedS is true when we have tried a prime greater or equal to sqrt(n)
//
function [ result , status ] = factor_memorylesstd ( n , imax )
    result = []
    // Get all primes up to sqrt(n)
    s = ceil ( sqrt ( n ) )
    u = n
    p = 2
    passedS = %f
    add2 = %t
    status = %f
    for i = 1 : imax
      number_log ( sprintf ( "Loop #%d/%d, u=%s\n" , i , imax , string ( u ) ) )
      if ( u == 1 ) then
        number_log ( sprintf ( "Found u=1, done.\n" ) )
        status = %t
        break
      end
      r = modulo ( u , p )
      if ( r==0 ) then
        number_log ( sprintf ( "Found divisor %s\n" , string( p )) )
        result ( 1 , $ + 1 ) = p
        u = u / p
      else
        // p does not divide u
        if ( ~passedS ) then
          // There are primes left to test
          if ( p == 2 ) then
            p = 3
          elseif ( p == 3 ) then
            p = 5
          elseif ( add2 ) then
            p = p + 2
            add2 = %f
          else
            p = p + 4
            add2 = %t
          end
          if ( p >= s ) then
            passedS = %t
          end
          number_log ( sprintf ( "Next divisor: %s\n" , string ( p ) ) )
        else
          // There are no primes left to test : the last number n is a prime
          number_log ( sprintf ( "No prime left to test, done.\n" ) )
          result ( 1 , $ + 1 ) = u
          status = %t
          break
        end
      end
    end
    if ( ~status ) then
      number_log ( sprintf ( "Failed to factor completely, u = %s.\n" , string ( u ) ) )
      result ( 1 , $ + 1 ) = u
    end
endfunction

